"""
測試 Advanced Operations 相關指令

涵蓋指令：
- groovy: 執行 Groovy script

測試重點：
- 執行簡單的 Groovy script
- 使用 script 查詢 Jenkins 資訊
- 從檔案載入 script
- 錯誤處理與安全性驗證
"""
from dataclasses import dataclass
from pathlib import Path


# ============================================================================
# Parser Functions - 將命令輸出解析為可精確比對的資料結構
# ============================================================================


@dataclass
class GroovyOutput:
    """Groovy script 輸出的結構化資料"""
    stdout: str
    stderr: str
    success: bool
    lines: list[str]

    @property
    def first_line(self) -> str:
        """取得第一行輸出（如果有的話）"""
        return self.lines[0] if self.lines else ""


def parse_groovy_output(result) -> GroovyOutput:
    """
    解析 groovy 命令的輸出

    預期格式：
        - stdout: script 的 println 輸出（每行一個結果）
        - stderr: 錯誤訊息（如果有）
        - returncode: 0 表示成功

    範例輸出：
        Hello from Groovy
        Line 2
        Line 3

    Returns:
        GroovyOutput: 結構化的輸出資料
    """
    stdout = result.stdout.strip()
    stderr = result.stderr.strip()
    success = result.returncode == 0

    # 將輸出分割成行，過濾空行
    lines = [line.strip() for line in stdout.split('\n') if line.strip()]

    return GroovyOutput(
        stdout=stdout,
        stderr=stderr,
        success=success,
        lines=lines
    )


# ============================================================================
# 測試函數 - 基本 Groovy 執行
# ============================================================================


def test_groovy_simple_hello(run_jenkee_authed, tmp_path):
    """
    測試執行簡單的 Hello World script

    對應 test plan 步驟 1
    """
    # Arrange: 準備簡單的 Hello World script
    script = "println('Hello from Groovy')"
    script_file = tmp_path / "hello.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證輸出內容
    assert "Hello from Groovy" in output.stdout, \
        f"Expected 'Hello from Groovy' in output, got: {output.stdout}"


def test_groovy_with_confirmation_cancelled(run_jenkee_authed, tmp_path):
    """
    測試取消執行 Groovy script（模擬輸入 n）

    對應文件中的「測試 2: 取消執行」
    """
    # Arrange: 準備簡單的 script
    script = "println('Hello from Groovy')"
    script_file = tmp_path / "hello-cancel.groovy"
    script_file.write_text(script)

    # Act: 嘗試執行但取消（模擬輸入 'n'）
    result = run_jenkee_authed.build_command(
        "groovy", str(script_file)
    ).with_stdin("n\n").run()

    # Assert: 驗證返回 0（取消不是錯誤）
    assert result.returncode == 0, f"Should return 0 when cancelled, got: {result.returncode}"
    assert "cancelled" in result.stdout.lower() or "canceled" in result.stdout.lower(), \
        "Should show cancellation message"


def test_groovy_with_confirmation_confirmed(run_jenkee_authed, tmp_path):
    """
    測試互動式確認後執行 Groovy script（模擬輸入 y）

    對應文件中的「測試 1: 互動式確認」
    """
    # Arrange: 準備簡單的 script
    script = "println('Hello from Groovy')"
    script_file = tmp_path / "hello-confirm.groovy"
    script_file.write_text(script)

    # Act: 執行並確認（模擬輸入 'y'）
    result = run_jenkee_authed.build_command(
        "groovy", str(script_file)
    ).with_stdin("y\n").run()

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert "Hello from Groovy" in result.stdout, \
        "Should show groovy script output after confirmation"


def test_groovy_jenkins_version(run_jenkee_authed, tmp_path):
    """
    測試查詢 Jenkins 版本

    對應 test plan 步驟 2
    """
    # Arrange: 準備查詢版本的 script
    script = """
import jenkins.model.Jenkins
println(Jenkins.instance.version)
"""
    script_file = tmp_path / "version.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證輸出是版本號格式（例如：2.440.1）
    version = output.first_line
    assert version, "Version should not be empty"
    # 驗證版本號格式（至少有數字和點）
    assert any(c.isdigit() for c in version), \
        f"Version should contain digits, got: {version}"


def test_groovy_list_jobs(run_jenkee_authed, tmp_path):
    """
    測試列出所有 Jobs

    對應 test plan 步驟 3
    """
    # Arrange: 準備列出 jobs 的 script
    script = """
import jenkins.model.Jenkins
Jenkins.instance.items.each { job ->
    println(job.name)
}
"""
    script_file = tmp_path / "list-jobs.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證至少列出了一些 jobs（測試環境有 test-job-1, test-job-2, test-job-3）
    job_names = set(output.lines)
    expected_jobs = {"test-job-1", "test-job-2", "test-job-3"}
    assert expected_jobs.issubset(job_names), \
        f"Expected {expected_jobs} to be in {job_names}"


def test_groovy_job_count(run_jenkee_authed, tmp_path):
    """
    測試計算 Jobs 數量

    對應 test plan 步驟 4
    """
    # Arrange: 準備計算 jobs 數量的 script
    script = """
import jenkins.model.Jenkins
println(Jenkins.instance.items.size())
"""
    script_file = tmp_path / "count-jobs.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證輸出是數字且大於 0
    count = output.first_line
    assert count.isdigit(), f"Count should be a number, got: {count}"
    assert int(count) >= 3, f"Should have at least 3 jobs, got: {count}"


def test_groovy_multiline_script(run_jenkee_authed, tmp_path):
    """
    測試執行多行 script

    對應 test plan 步驟 6
    """
    # Arrange: 準備多行 script
    script = """
import jenkins.model.Jenkins

def jenkins = Jenkins.instance

println("=== Jenkins Statistics ===")
println("Version: ${jenkins.version}")
println("Total Jobs: ${jenkins.items.size()}")
"""
    script_file = tmp_path / "multiline.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證輸出包含預期的標題和內容
    assert "=== Jenkins Statistics ===" in output.stdout, \
        "Should contain statistics header"
    assert "Version:" in output.stdout, "Should contain version info"
    assert "Total Jobs:" in output.stdout, "Should contain job count"


def test_groovy_query_specific_job(run_jenkee_authed, tmp_path):
    """
    測試查詢特定 Job 的資訊

    對應 test plan 步驟 5
    """
    # Arrange: 準備查詢特定 job 的 script
    script = """
import jenkins.model.Jenkins

def job = Jenkins.instance.getItem("test-job-1")
if (job != null) {
    println("Job: ${job.name}")
    println("Enabled: ${!job.isDisabled()}")
    println("Buildable: ${job.isBuildable()}")
} else {
    println("Job not found")
}
"""
    script_file = tmp_path / "query-job.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證輸出包含 job 資訊
    assert "Job: test-job-1" in output.stdout, "Should contain job name"
    assert "Enabled:" in output.stdout, "Should contain enabled status"
    assert "Buildable:" in output.stdout, "Should contain buildable status"


# ============================================================================
# 測試函數 - 錯誤情境
# ============================================================================


def test_groovy_syntax_error(run_jenkee_authed, tmp_path):
    """
    測試執行有語法錯誤的 script

    對應 test plan 錯誤情境測試
    """
    # Arrange: 準備有語法錯誤的 script
    script = "invalid groovy syntax {"
    script_file = tmp_path / "invalid.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令並允許失敗
    result = run_jenkee_authed.build_command(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    ).allow_failure().run()

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證失敗
    assert result.returncode != 0, "Should fail with syntax error"
    assert not output.success, "Script execution should fail"

    # Verify: 驗證錯誤訊息（可能在 stdout 或 stderr）
    error_output = (result.stdout + result.stderr).lower()
    # Jenkins groovy console 可能會在 stdout 或 stderr 顯示錯誤
    assert 'error' in error_output or 'exception' in error_output or 'unexpected' in error_output, \
        f"Should have error message, got stdout: {result.stdout}, stderr: {result.stderr}"


def test_groovy_null_handling(run_jenkee_authed, tmp_path):
    """
    測試存取不存在的 Job（null 處理）

    對應 test plan 錯誤情境測試
    """
    # Arrange: 準備查詢不存在 job 的 script（包含 null 檢查）
    script = """
import jenkins.model.Jenkins

def job = Jenkins.instance.getItem("non-existent-job")
if (job == null) {
    println("Job not found")
} else {
    println("Job found: ${job.name}")
}
"""
    script_file = tmp_path / "null-check.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功（script 本身沒問題，只是 job 不存在）
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證正確處理 null
    assert "Job not found" in output.stdout, \
        f"Should handle null correctly, got: {output.stdout}"


def test_groovy_empty_script(run_jenkee_authed, tmp_path):
    """
    測試執行空的 script

    對應 test plan 錯誤情境測試
    """
    # Arrange: 準備空的 script
    script = ""
    script_file = tmp_path / "empty.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令並允許失敗
    result = run_jenkee_authed.build_command(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    ).allow_failure().run()

    # Assert: 驗證失敗（empty script 應該被拒絕）
    assert result.returncode != 0, "Should fail with empty script"

    # Verify: 驗證錯誤訊息
    error_output = (result.stdout + result.stderr).lower()
    assert 'error' in error_output or 'empty' in error_output, \
        f"Should have error about empty script, got: {result.stdout + result.stderr}"


def test_groovy_nonexistent_file(run_jenkee_authed, tmp_path):
    """
    測試執行不存在的 script 檔案

    對應 test plan 錯誤情境測試
    """
    # Arrange: 使用不存在的檔案路徑
    script_file = tmp_path / "nonexistent.groovy"

    # Act: 執行 groovy 指令並允許失敗
    result = run_jenkee_authed.build_command(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    ).allow_failure().run()

    # Assert: 驗證失敗
    assert result.returncode != 0, "Should fail when file doesn't exist"

    # Verify: 驗證錯誤訊息
    error_output = (result.stdout + result.stderr).lower()
    assert 'error' in error_output or 'not found' in error_output or 'no such' in error_output, \
        f"Should have error about missing file, got: {result.stdout + result.stderr}"


# ============================================================================
# 測試函數 - 安全性與只讀操作
# ============================================================================


def test_groovy_read_only_operations(run_jenkee_authed, tmp_path):
    """
    測試只讀操作（查詢資訊不修改）

    驗證可以安全執行查詢操作
    """
    # Arrange: 準備只讀的查詢 script
    script = """
import jenkins.model.Jenkins

def jenkins = Jenkins.instance

// 只讀操作：查詢資訊
println("URL: ${jenkins.rootUrl ?: 'Not set'}")
println("Version: ${jenkins.version}")
println("Jobs: ${jenkins.items.size()}")
println("Nodes: ${jenkins.nodes.size() + 1}")  // +1 for master
"""
    script_file = tmp_path / "read-only.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證輸出包含所有查詢項目
    assert "URL:" in output.stdout, "Should contain URL info"
    assert "Version:" in output.stdout, "Should contain version info"
    assert "Jobs:" in output.stdout, "Should contain jobs count"
    assert "Nodes:" in output.stdout, "Should contain nodes count"


# ============================================================================
# 測試函數 - Dry-Run 模式範例
# ============================================================================


def test_groovy_dry_run_mode(run_jenkee_authed, tmp_path):
    """
    測試 Dry-Run 模式（不實際修改）

    對應 test plan 步驟 7
    """
    # Arrange: 準備 dry-run 模式的 script
    script = """
import jenkins.model.Jenkins

def dryRun = true  // 設為 true，不實際執行

Jenkins.instance.items.findAll { it.disabled }.each { job ->
    if (dryRun) {
        println("[DRY-RUN] Would enable job: ${job.name}")
    } else {
        job.enable()
        println("[EXECUTED] Enabled job: ${job.name}")
    }
}

if (dryRun) {
    println("This was a dry-run.")
}
"""
    script_file = tmp_path / "dry-run.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證是 dry-run 模式
    assert "This was a dry-run." in output.stdout, \
        "Should indicate dry-run mode"

    # Verify: 確認沒有實際執行（沒有 EXECUTED 訊息）
    assert "[EXECUTED]" not in output.stdout, \
        "Should not have executed any actual changes"


# ============================================================================
# 整合測試 - 完整工作流程
# ============================================================================


def test_groovy_batch_query_workflow(run_jenkee_authed, tmp_path):
    """
    測試完整的批次查詢工作流程

    對應 test plan 場景 A
    """
    # Arrange: 準備批次查詢 script
    script = """
import jenkins.model.Jenkins

def jenkins = Jenkins.instance

println("=== Jenkins Summary ===")
println("Version: ${jenkins.version}")
println("Total Jobs: ${jenkins.items.size()}")

println("")
println("=== Job List ===")
jenkins.items.each { job ->
    def lastBuild = job.lastBuild
    def status = lastBuild ? lastBuild.result : "Never built"
    println("${job.name}: ${status}")
}
"""
    script_file = tmp_path / "batch-query.groovy"
    script_file.write_text(script)

    # Act: 執行 groovy 指令
    result = run_jenkee_authed.run(
        "groovy", str(script_file), "--yes-i-really-mean-it"
    )

    # Parse: 解析輸出
    output = parse_groovy_output(result)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"groovy should succeed, got: {result.stderr}"
    assert output.success, "Script execution should succeed"

    # Verify: 驗證輸出結構完整
    assert "=== Jenkins Summary ===" in output.stdout, \
        "Should contain summary section"
    assert "=== Job List ===" in output.stdout, \
        "Should contain job list section"
    assert "Version:" in output.stdout, "Should contain version"
    assert "Total Jobs:" in output.stdout, "Should contain job count"

    # Verify: 驗證至少列出了測試 jobs
    assert "test-job-1" in output.stdout, "Should list test-job-1"
    assert "test-job-2" in output.stdout, "Should list test-job-2"



"""
測試 Job Organization and Status Management 相關指令

涵蓋指令：
- job-status: 查看 job 狀態與觸發關係
- add-job-to-view: 將 jobs 加入 view
- enable-job: 啟用 job
- disable-job: 停用 job

測試重點：
- 驗證 job 狀態查詢功能
- 測試將 jobs 組織到 views
- 測試啟用/停用 job 的完整流程
- 驗證冪等性與批次操作
"""
import re
from dataclasses import dataclass
from typing import List, Optional, Set


# ============================================================================
# Parser Functions - 將命令輸出解析為可精確比對的資料結構
# ============================================================================


@dataclass
class JobStatusInfo:
    """Job 狀態的結構化資料"""
    name: str
    status: str  # "ENABLED" or "DISABLED"
    buildable: bool
    last_build: Optional[int] = None
    last_stable_build: Optional[int] = None
    last_successful_build: Optional[int] = None
    last_failed_build: Optional[int] = None
    downstream_projects: List[str] = None
    upstream_projects: List[str] = None

    def __post_init__(self):
        if self.downstream_projects is None:
            self.downstream_projects = []
        if self.upstream_projects is None:
            self.upstream_projects = []


def parse_job_status(stdout: str) -> JobStatusInfo:
    """
    解析 job-status 命令的輸出

    預期格式：
        === Job: test-job-1 ===

        Status: ENABLED
        Buildable: true

        === Health ===
        ...

        === Last Builds ===
        Last Build: #5
        Last Stable Build: #4
        ...

        === Downstream Projects ===
          - downstream-job-1
          (none)

        === Upstream Projects ===
          - upstream-job-1
          (none)

    Returns:
        JobStatusInfo: 結構化的 job 狀態資料
    """
    lines = stdout.split('\n')

    name = None
    status = None
    buildable = False
    last_build = None
    last_stable_build = None
    last_successful_build = None
    last_failed_build = None
    downstream_projects = []
    upstream_projects = []

    current_section = None

    for line in lines:
        line_stripped = line.strip()

        # Job header
        if line_stripped.startswith('=== Job:'):
            match = re.search(r'===\s*Job:\s*(.+?)\s*===', line_stripped)
            if match:
                name = match.group(1)

        # Status
        elif line_stripped.startswith('Status:'):
            status = line_stripped.split(':', 1)[1].strip()

        # Buildable
        elif line_stripped.startswith('Buildable:'):
            buildable_str = line_stripped.split(':', 1)[1].strip().lower()
            buildable = buildable_str == 'true'

        # Section headers
        elif '=== Last Builds ===' in line_stripped:
            current_section = 'builds'
        elif '=== Downstream Projects ===' in line_stripped:
            current_section = 'downstream'
        elif '=== Upstream Projects ===' in line_stripped:
            current_section = 'upstream'

        # Last builds
        elif current_section == 'builds':
            if line_stripped.startswith('Last Build:'):
                match = re.search(r'#(\d+)', line_stripped)
                if match:
                    last_build = int(match.group(1))
            elif line_stripped.startswith('Last Stable Build:'):
                match = re.search(r'#(\d+)', line_stripped)
                if match:
                    last_stable_build = int(match.group(1))
            elif line_stripped.startswith('Last Successful Build:'):
                match = re.search(r'#(\d+)', line_stripped)
                if match:
                    last_successful_build = int(match.group(1))
            elif line_stripped.startswith('Last Failed Build:'):
                match = re.search(r'#(\d+)', line_stripped)
                if match:
                    last_failed_build = int(match.group(1))

        # Downstream projects
        elif current_section == 'downstream':
            if line_stripped.startswith('- '):
                project = line_stripped[2:].strip()
                if project and project != '(none)':
                    downstream_projects.append(project)

        # Upstream projects
        elif current_section == 'upstream':
            if line_stripped.startswith('- '):
                project = line_stripped[2:].strip()
                if project and project != '(none)':
                    upstream_projects.append(project)

    return JobStatusInfo(
        name=name or '',
        status=status or 'UNKNOWN',
        buildable=buildable,
        last_build=last_build,
        last_stable_build=last_stable_build,
        last_successful_build=last_successful_build,
        last_failed_build=last_failed_build,
        downstream_projects=downstream_projects,
        upstream_projects=upstream_projects
    )


def parse_jobs_list(stdout: str) -> Set[str]:
    """
    解析 list-jobs 命令的輸出

    預期格式：每行一個 job 名稱
        test-job-1
        test-job-2
        test-job-3

    Returns:
        Set[str]: job 名稱的集合
    """
    lines = stdout.strip().split('\n')
    jobs = set()

    for line in lines:
        line = line.strip()
        if line and not line.startswith('==='):  # 跳過空行與分隔線
            jobs.add(line)

    return jobs


# ============================================================================
# 測試函數 - Job-Status 指令
# ============================================================================


def test_job_status_basic(run_jenkee_authed):
    """
    測試查看 Job 狀態

    對應 test plan 步驟 1
    """
    # Arrange: 使用已認證的 jenkee runner（由 fixture 提供）

    # Act: 執行 job-status 指令
    result = run_jenkee_authed.run("job-status", "test-job-1")

    # Parse: 解析 job 狀態
    status = parse_job_status(result.stdout)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"job-status should succeed, got: {result.stderr}"
    assert status.name == "test-job-1", f"Expected job name test-job-1, got {status.name}"
    assert status.status in ["ENABLED", "DISABLED"], \
        f"Status should be ENABLED or DISABLED, got {status.status}"


def test_job_status_shows_build_info(run_jenkee_authed):
    """
    測試 job-status 顯示 Build 資訊

    驗證輸出包含 last build 相關資訊
    """
    # Arrange: 先觸發一個 build 確保有 build history
    build_result = run_jenkee_authed.run("build", "test-job-1")
    assert build_result.returncode == 0, "Should trigger build successfully"

    # Act: 執行 job-status 指令
    result = run_jenkee_authed.run("job-status", "test-job-1")

    # Parse: 解析 job 狀態
    status = parse_job_status(result.stdout)

    # Assert: 驗證包含 build 資訊
    assert result.returncode == 0
    assert status.last_build is not None, "Should have last build information"


def test_job_status_nonexistent(run_jenkee_authed):
    """
    測試查看不存在的 Job 狀態

    對應 test plan 錯誤情境測試
    """
    # Arrange: 使用不存在的 job 名稱

    # Act: 執行 job-status 指令並允許失敗
    result = run_jenkee_authed.build_command(
        "job-status", "non-existent-job"
    ).allow_failure().run()

    # Assert: 驗證失敗
    assert result.returncode != 0, "Should fail for non-existent job"
    error_output = (result.stdout + result.stderr).lower()
    assert 'error' in error_output or 'not found' in error_output, \
        f"Should have error message, got: {result.stdout + result.stderr}"


# ============================================================================
# 測試函數 - Add-Job-To-View 指令
# ============================================================================


def test_add_job_to_view_single(run_jenkee_authed):
    """
    測試將單一 Job 加入 View

    對應 test plan 步驟 2
    """
    # Arrange: 使用已存在的 test-view 和 test-job-3
    # test-view 已經包含 test-job-1 和 test-job-2，我們加入 test-job-3

    # Act: 執行 add-job-to-view 指令
    result = run_jenkee_authed.run("add-job-to-view", "test-view", "test-job-3")

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"add-job-to-view should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"


def test_verify_job_in_view(run_jenkee_authed):
    """
    測試驗證 Job 已加入 View

    對應 test plan 步驟 3
    """
    # Arrange: 先將 job 加入 view
    add_result = run_jenkee_authed.run("add-job-to-view", "test-view", "test-job-3")
    assert add_result.returncode == 0, "Should add job successfully"

    # Act: 執行 list-jobs 指令查看 view 中的 jobs
    result = run_jenkee_authed.run("list-jobs", "test-view")

    # Parse: 解析 jobs 列表
    jobs = parse_jobs_list(result.stdout)

    # Assert: 驗證 job 存在於 view 中
    assert result.returncode == 0
    assert "test-job-3" in jobs, \
        f"test-job-3 should be in test-view, got: {jobs}"


def test_add_job_to_view_multiple(run_jenkee_authed):
    """
    測試批次將多個 Jobs 加入 View

    對應 test plan 步驟 4
    """
    # Arrange: 使用已存在的 test-view
    # test-view 已經包含 test-job-1 和 test-job-2
    # 我們加入 test-job-3 和 long-running-job

    # Act: 執行 add-job-to-view 指令（批次）
    result = run_jenkee_authed.run(
        "add-job-to-view", "test-view", "test-job-3", "long-running-job"
    )

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"add-job-to-view should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"

    # Verify: 驗證所有 jobs 都加入了
    list_result = run_jenkee_authed.run("list-jobs", "test-view")
    jobs = parse_jobs_list(list_result.stdout)

    assert "test-job-3" in jobs, "test-job-3 should be in view"
    assert "long-running-job" in jobs, "long-running-job should be in view"


def test_add_job_to_view_idempotent(run_jenkee_authed):
    """
    測試冪等性（重複加入相同 Job）

    對應 test plan 步驟 5
    """
    # Arrange: 先加入一次（test-job-1 在 test-view 可能已經存在）
    first_result = run_jenkee_authed.run("add-job-to-view", "test-view", "test-job-1")
    assert first_result.returncode == 0, "First add should succeed"

    # Act: 再次加入相同的 job（測試冪等性）
    second_result = run_jenkee_authed.run("add-job-to-view", "test-view", "test-job-1")

    # Assert: 驗證操作成功（冪等）
    assert second_result.returncode == 0, \
        "Second add should also succeed (idempotent operation)"

    # Verify: 驗證 job 只出現一次
    list_result = run_jenkee_authed.run("list-jobs", "test-view")
    jobs_list = list_result.stdout.strip().split('\n')
    test_job_1_count = sum(1 for job in jobs_list if job.strip() == "test-job-1")

    assert test_job_1_count == 1, \
        f"test-job-1 should appear exactly once, but appeared {test_job_1_count} times"


def test_add_job_to_nonexistent_view(run_jenkee_authed):
    """
    測試將 Job 加入不存在的 View

    對應 test plan 錯誤情境測試
    """
    # Arrange: 使用不存在的 view 名稱

    # Act: 執行 add-job-to-view 指令並允許失敗
    result = run_jenkee_authed.build_command(
        "add-job-to-view", "non-existent-view", "test-job-1"
    ).allow_failure().run()

    # Assert: 驗證失敗
    assert result.returncode != 0, "Should fail for non-existent view"
    error_output = (result.stdout + result.stderr).lower()
    assert 'error' in error_output or 'not found' in error_output or 'no such' in error_output, \
        f"Should have error message, got: {result.stdout + result.stderr}"


# ============================================================================
# 測試函數 - Enable/Disable Job 指令
# ============================================================================


def test_disable_job(run_jenkee_authed):
    """
    測試停用 Job

    對應 test plan 步驟 6
    注意：測試後會恢復 job 狀態
    """
    # Arrange: 確保 test-job-3 初始是啟用的
    enable_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert enable_result.returncode == 0, "Should enable job first"

    # Act: 執行 disable-job 指令
    result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"disable-job should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"

    # Cleanup: 恢復 job 狀態（啟用）
    cleanup_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert cleanup_result.returncode == 0, "Should restore job state"


def test_disable_job_with_confirmation_cancelled(run_jenkee_authed):
    """
    測試取消停用 Job（模擬輸入 n）

    對應文件中的「測試 2: 取消停用」
    """
    # Arrange: 確保 test-job-3 初始是啟用的
    enable_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert enable_result.returncode == 0, "Should enable job first"

    # Act: 嘗試停用但取消（模擬輸入 'n'）
    result = run_jenkee_authed.build_command(
        "disable-job", "test-job-3"
    ).with_stdin("n\n").run()

    # Assert: 驗證返回 0（取消不是錯誤）
    assert result.returncode == 0, f"Should return 0 when cancelled, got: {result.returncode}"
    assert "cancelled" in result.stdout.lower() or "canceled" in result.stdout.lower(), \
        "Should show cancellation message"

    # Verify: job 仍然是啟用狀態
    status_result = run_jenkee_authed.run("job-status", "test-job-3")
    status = parse_job_status(status_result.stdout)
    assert status.status == "ENABLED", "Job should remain enabled after cancellation"


def test_disable_job_with_confirmation_confirmed(run_jenkee_authed):
    """
    測試互動式確認後停用 Job（模擬輸入 y）

    對應文件中的「測試 1: 互動式確認」
    """
    # Arrange: 確保 test-job-3 初始是啟用的
    enable_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert enable_result.returncode == 0, "Should enable job first"

    # Act: 停用並確認（模擬輸入 'y'）
    result = run_jenkee_authed.build_command(
        "disable-job", "test-job-3"
    ).with_stdin("y\n").run()

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"disable-job should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"

    # Verify: job 已停用
    status_result = run_jenkee_authed.run("job-status", "test-job-3")
    status = parse_job_status(status_result.stdout)
    assert status.status == "DISABLED", "Job should be disabled after confirmation"

    # Cleanup: 恢復 job 狀態（啟用）
    cleanup_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert cleanup_result.returncode == 0, "Should restore job state"


def test_verify_job_disabled(run_jenkee_authed):
    """
    測試驗證 Job 已停用

    對應 test plan 步驟 7
    """
    # Arrange: 先停用 job
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Should disable job successfully"

    # Act: 執行 job-status 指令查看狀態
    result = run_jenkee_authed.run("job-status", "test-job-3")

    # Parse: 解析 job 狀態
    status = parse_job_status(result.stdout)

    # Assert: 驗證狀態為 DISABLED
    assert result.returncode == 0
    assert status.status == "DISABLED", \
        f"Job should be DISABLED, got {status.status}"

    # Cleanup: 恢復 job 狀態
    cleanup_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert cleanup_result.returncode == 0, "Should restore job state"


def test_build_disabled_job(run_jenkee_authed):
    """
    測試嘗試觸發已停用的 Job

    對應 test plan 步驟 8
    """
    # Arrange: 先停用 job
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Should disable job successfully"

    # Act: 嘗試觸發 build 並允許失敗
    result = run_jenkee_authed.build_command("build", "test-job-3").allow_failure().run()

    # Assert: 驗證失敗或顯示警告
    # 注意：Jenkins 的行為可能是失敗或成功但不執行
    # 我們只驗證有適當的訊息
    output = (result.stdout + result.stderr).lower()
    # 可能的情況：失敗、或成功但提示 job 已停用
    assert result.returncode != 0 or 'disabled' in output or 'not buildable' in output, \
        "Should fail or warn about disabled job"

    # Cleanup: 恢復 job 狀態
    cleanup_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert cleanup_result.returncode == 0, "Should restore job state"


def test_enable_job(run_jenkee_authed):
    """
    測試重新啟用 Job

    對應 test plan 步驟 9
    """
    # Arrange: 先停用 job
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Should disable job first"

    # Act: 執行 enable-job 指令
    result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"enable-job should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"


def test_enable_job_with_confirmation_cancelled(run_jenkee_authed):
    """
    測試取消啟用 Job（模擬輸入 n）

    對應文件中的「測試 2: 取消啟用」
    """
    # Arrange: 確保 test-job-3 初始是停用的
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Should disable job first"

    # Act: 嘗試啟用但取消（模擬輸入 'n'）
    result = run_jenkee_authed.build_command(
        "enable-job", "test-job-3"
    ).with_stdin("n\n").run()

    # Assert: 驗證返回 0（取消不是錯誤）
    assert result.returncode == 0, f"Should return 0 when cancelled, got: {result.returncode}"
    assert "cancelled" in result.stdout.lower() or "canceled" in result.stdout.lower(), \
        "Should show cancellation message"

    # Verify: job 仍然是停用狀態
    status_result = run_jenkee_authed.run("job-status", "test-job-3")
    status = parse_job_status(status_result.stdout)
    assert status.status == "DISABLED", "Job should remain disabled after cancellation"


def test_enable_job_with_confirmation_confirmed(run_jenkee_authed):
    """
    測試互動式確認後啟用 Job（模擬輸入 y）

    對應文件中的「測試 1: 互動式確認」
    """
    # Arrange: 確保 test-job-3 初始是停用的
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Should disable job first"

    # Act: 啟用並確認（模擬輸入 'y'）
    result = run_jenkee_authed.build_command(
        "enable-job", "test-job-3"
    ).with_stdin("y\n").run()

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"enable-job should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"

    # Verify: job 已啟用
    status_result = run_jenkee_authed.run("job-status", "test-job-3")
    status = parse_job_status(status_result.stdout)
    assert status.status == "ENABLED", "Job should be enabled after confirmation"


def test_verify_job_enabled(run_jenkee_authed):
    """
    測試驗證 Job 已啟用

    對應 test plan 步驟 10
    """
    # Arrange: 先停用再啟用 job
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Should disable job first"

    enable_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert enable_result.returncode == 0, "Should enable job successfully"

    # Act: 執行 job-status 指令查看狀態
    result = run_jenkee_authed.run("job-status", "test-job-3")

    # Parse: 解析 job 狀態
    status = parse_job_status(result.stdout)

    # Assert: 驗證狀態為 ENABLED
    assert result.returncode == 0
    assert status.status == "ENABLED", \
        f"Job should be ENABLED, got {status.status}"

    # Verify: 驗證可以觸發 build
    build_result = run_jenkee_authed.run("build", "test-job-3")
    assert build_result.returncode == 0, "Should be able to trigger build"


def test_disable_multiple_jobs(run_jenkee_authed):
    """
    測試批次停用多個 Jobs

    對應 test plan 步驟 11
    """
    # Arrange: 確保 jobs 初始是啟用的
    enable_result = run_jenkee_authed.run(
        "enable-job", "test-job-1", "test-job-2", "--yes-i-really-mean-it"
    )
    assert enable_result.returncode == 0, "Should enable jobs first"

    # Act: 執行 disable-job 指令（批次）
    result = run_jenkee_authed.run(
        "disable-job", "test-job-1", "test-job-2", "--yes-i-really-mean-it"
    )

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"disable-job should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"

    # Verify: 驗證所有 jobs 都被停用
    status1_result = run_jenkee_authed.run("job-status", "test-job-1")
    status1 = parse_job_status(status1_result.stdout)
    assert status1.status == "DISABLED", "test-job-1 should be disabled"

    status2_result = run_jenkee_authed.run("job-status", "test-job-2")
    status2 = parse_job_status(status2_result.stdout)
    assert status2.status == "DISABLED", "test-job-2 should be disabled"

    # Cleanup: 恢復 jobs 狀態
    cleanup_result = run_jenkee_authed.run(
        "enable-job", "test-job-1", "test-job-2", "--yes-i-really-mean-it"
    )
    assert cleanup_result.returncode == 0, "Should restore jobs state"


def test_enable_multiple_jobs(run_jenkee_authed):
    """
    測試批次啟用多個 Jobs

    對應 test plan 步驟 12
    """
    # Arrange: 先停用 jobs
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-1", "test-job-2", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Should disable jobs first"

    # Act: 執行 enable-job 指令（批次）
    result = run_jenkee_authed.run(
        "enable-job", "test-job-1", "test-job-2", "--yes-i-really-mean-it"
    )

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"enable-job should succeed, got: {result.stderr}"
    assert "✓" in result.stdout or "success" in result.stdout.lower(), \
        "Should show success message"

    # Verify: 驗證所有 jobs 都被啟用
    status1_result = run_jenkee_authed.run("job-status", "test-job-1")
    status1 = parse_job_status(status1_result.stdout)
    assert status1.status == "ENABLED", "test-job-1 should be enabled"

    status2_result = run_jenkee_authed.run("job-status", "test-job-2")
    status2 = parse_job_status(status2_result.stdout)
    assert status2.status == "ENABLED", "test-job-2 should be enabled"


def test_disable_nonexistent_job(run_jenkee_authed):
    """
    測試停用不存在的 Job

    對應 test plan 錯誤情境測試
    """
    # Arrange: 使用不存在的 job 名稱

    # Act: 執行 disable-job 指令並允許失敗
    result = run_jenkee_authed.build_command(
        "disable-job", "non-existent-job", "--yes-i-really-mean-it"
    ).allow_failure().run()

    # Assert: 驗證失敗
    assert result.returncode != 0, "Should fail for non-existent job"
    error_output = (result.stdout + result.stderr).lower()
    assert 'error' in error_output or 'not found' in error_output or 'no such' in error_output, \
        f"Should have error message, got: {result.stdout + result.stderr}"


# ============================================================================
# 整合測試 - 完整工作流程
# ============================================================================


def test_complete_job_organization_workflow(run_jenkee_authed):
    """
    測試完整的 job 組織工作流程

    對應 test plan 場景 A：組織相關 jobs 到專案 view
    """
    # 1. 列出所有 jobs
    list_result = run_jenkee_authed.run("list-jobs", "--all")
    all_jobs = parse_jobs_list(list_result.stdout)

    assert list_result.returncode == 0, "Step 1: list-jobs should succeed"
    assert len(all_jobs) > 0, "Step 1: Should have at least one job"

    # 2. 批次加入 jobs 到 view（test-view 已經有 job，測試冪等性）
    add_result = run_jenkee_authed.run(
        "add-job-to-view", "test-view", "test-job-1", "test-job-2"
    )

    assert add_result.returncode == 0, "Step 2: add-job-to-view should succeed"

    # 3. 驗證結果
    verify_result = run_jenkee_authed.run("list-jobs", "test-view")
    view_jobs = parse_jobs_list(verify_result.stdout)

    assert verify_result.returncode == 0, "Step 3: list-jobs should succeed"
    assert "test-job-1" in view_jobs, "Step 3: test-job-1 should be in view"
    assert "test-job-2" in view_jobs, "Step 3: test-job-2 should be in view"

    # 4. 查看 jobs 狀態
    for job in ["test-job-1", "test-job-2"]:
        status_result = run_jenkee_authed.run("job-status", job)
        assert status_result.returncode == 0, f"Step 4: job-status should succeed for {job}"


def test_enable_disable_workflow(run_jenkee_authed):
    """
    測試啟用/停用完整流程

    對應 test plan 場景 B：維護期間暫停 jobs
    驗證 enable/disable 是可逆操作
    """
    # 1. 確認初始狀態（啟用）
    initial_status_result = run_jenkee_authed.run("job-status", "test-job-3")
    initial_status = parse_job_status(initial_status_result.stdout)

    # 如果初始是 disabled，先啟用
    if initial_status.status == "DISABLED":
        enable_result = run_jenkee_authed.run(
            "enable-job", "test-job-3", "--yes-i-really-mean-it"
        )
        assert enable_result.returncode == 0, "Should enable job first"

    # 2. 停用 job
    disable_result = run_jenkee_authed.run(
        "disable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert disable_result.returncode == 0, "Step 2: disable-job should succeed"

    # 3. 驗證已停用
    disabled_status_result = run_jenkee_authed.run("job-status", "test-job-3")
    disabled_status = parse_job_status(disabled_status_result.stdout)
    assert disabled_status.status == "DISABLED", \
        "Step 3: Job should be disabled"

    # 4. 重新啟用
    enable_result = run_jenkee_authed.run(
        "enable-job", "test-job-3", "--yes-i-really-mean-it"
    )
    assert enable_result.returncode == 0, "Step 4: enable-job should succeed"

    # 5. 驗證已啟用
    enabled_status_result = run_jenkee_authed.run("job-status", "test-job-3")
    enabled_status = parse_job_status(enabled_status_result.stdout)
    assert enabled_status.status == "ENABLED", \
        "Step 5: Job should be enabled"

    # 6. 驗證可以觸發 build
    build_result = run_jenkee_authed.run("build", "test-job-3")
    assert build_result.returncode == 0, \
        "Step 6: Should be able to trigger build after enabling"


def test_job_organization_output_is_parseable(run_jenkee_authed):
    """
    測試 job organization 指令的輸出可以被腳本解析

    確保輸出格式穩定，適合自動化使用
    """
    # Test job-status
    status_result = run_jenkee_authed.run("job-status", "test-job-1")
    status = parse_job_status(status_result.stdout)

    # 應該能夠成功解析
    assert status.name == "test-job-1", "Should be able to parse job-status output"
    assert status.status in ["ENABLED", "DISABLED"], "Should have valid status"

    # Test list-jobs (from view)
    list_result = run_jenkee_authed.run("list-jobs", "test-view")
    jobs = parse_jobs_list(list_result.stdout)

    # 應該能夠成功解析
    assert len(jobs) > 0, "Should be able to parse list-jobs output"
    assert all(isinstance(job, str) for job in jobs), "All jobs should be strings"



"""Helper test to view Jenkins logs and init script execution results"""


def test_view_jenkins_init_output(jenkins_instance):
    """
    查看 Jenkins 初始化輸出（包含 fixture scripts 的執行結果）

    這個測試主要用於 debug 和驗證 fixture scripts 是否正確執行。
    執行時使用 -s flag 來查看輸出：pytest tests/test_jenkins_logs_helper.py::test_view_jenkins_init_output -s
    """
    # 方式 1：取得分離的 stdout 和 stderr
    stdout, stderr = jenkins_instance.get_logs()

    # 方式 2：取得合併的 logs（不標示 stream）
    combined_logs = jenkins_instance.get_logs_combined()

    # 方式 3：取得合併的 logs（標示 stdout/stderr）
    combined_logs_marked = jenkins_instance.get_logs_combined(mark_streams=True)

    print("\n" + "=" * 80)
    print("Jenkins Init Scripts Execution Output")
    print("=" * 80)

    # 過濾出我們關心的訊息
    for line in combined_logs.split('\n'):
        if any(keyword in line for keyword in [
            'Created job',
            'Created view',
            'Added',
            'Verification',
            'Test jobs and views setup completed',
            'GroovyHookScript#execute',
            '01-create-test-jobs.groovy'
        ]):
            print(line)

    print("=" * 80)

    # 驗證關鍵訊息存在
    assert 'Created job: test-job-1' in combined_logs
    assert 'Created job: test-job-2' in combined_logs
    assert 'Created job: test-job-3' in combined_logs
    assert 'Created view: test-view' in combined_logs
    assert 'Created empty view: empty-view' in combined_logs


def test_separate_stdout_stderr(jenkins_instance):
    """
    示範如何分別取得 stdout 和 stderr

    有時候需要分別查看 stdout 和 stderr 來除錯。
    """
    stdout, stderr = jenkins_instance.get_logs()

    # 處理可能是 bytes 或 str 的情況
    stdout_text = stdout.decode('utf-8', errors='ignore') if isinstance(stdout, bytes) else stdout
    stderr_text = stderr.decode('utf-8', errors='ignore') if isinstance(stderr, bytes) else stderr

    print("\n=== STDOUT ===")
    print(f"Length: {len(stdout_text)} characters")

    print("\n=== STDERR ===")
    print(f"Length: {len(stderr_text)} characters")

    # Jenkins 的主要輸出通常在 stdout
    assert len(stdout_text) > 0 or len(stderr_text) > 0, "Should have some logs"



"""Update job command"""

import sys

from jenkins_tools.core import Command, JenkinsConfig, JenkinsCLI


class UpdateJobCommand(Command):
    """Update an existing Jenkins job configuration from XML"""

    def __init__(self, args):
        """
        Initialize with command line arguments

        Args:
            args: List of command arguments (sys.argv[2:])
                  First argument is job name
        """
        self.args = args

    def execute(self) -> int:
        """Execute update-job command"""
        config = JenkinsConfig()

        # Check if credentials are configured
        if not config.is_configured():
            print("Error: Jenkins credentials not configured.", file=sys.stderr)
            print(f"Run 'jks auth' to configure credentials.", file=sys.stderr)
            return 1

        # Parse arguments
        if len(self.args) != 1:
            print("Error: Job name is required.", file=sys.stderr)
            print("Usage: jks update-job <job-name> < config.xml", file=sys.stderr)
            return 1

        job_name = self.args[0]

        # Read XML from stdin
        if sys.stdin.isatty():
            print("Error: No XML input provided.", file=sys.stderr)
            print("Usage: jks update-job <job-name> < config.xml", file=sys.stderr)
            print("   or: jks get-job <job> | jks update-job <job>", file=sys.stderr)
            return 1

        xml_content = sys.stdin.read()

        if not xml_content.strip():
            print("Error: XML input is empty.", file=sys.stderr)
            return 1

        # Use Jenkins CLI to update job
        cli = JenkinsCLI(config)
        result = cli.run("update-job", job_name, stdin_input=xml_content)

        if result.returncode != 0:
            print(f"Error: Failed to update job '{job_name}'", file=sys.stderr)
            if result.stderr:
                print(result.stderr, file=sys.stderr)
            return 1

        # Success
        print(f"✓ Successfully updated job '{job_name}'")

        return 0



"""
測試 Credentials Management 相關指令

涵蓋指令：
- list-credentials: 列出所有 credentials metadata
- describe-credentials: 查看特定 credential 詳細資訊

測試重點：
- 驗證不洩漏 secrets（在沒有 --show-secret 參數時）
- 測試不同類型的 credentials
- 結構化輸出解析
- Read-only 操作，安全性高
"""
import re
from dataclasses import dataclass
from typing import List, Optional


# ============================================================================
# Parser Functions - 將命令輸出解析為可精確比對的資料結構
# ============================================================================


@dataclass
class Credential:
    """單一 credential 的資訊"""
    id: str
    type: str
    scope: Optional[str] = None
    description: Optional[str] = None
    username: Optional[str] = None  # 只適用於 UsernamePassword 類型


@dataclass
class CredentialsDomain:
    """Credentials domain 的資訊"""
    name: str
    credentials: List[Credential]


def parse_credentials_list(stdout: str) -> List[CredentialsDomain]:
    """
    解析 list-credentials 命令的輸出

    預期格式：
        === Domain: (global) ===

        ID: test-credential-1
          Type: UsernamePasswordCredentialsImpl
          Scope: GLOBAL
          Description: Test username/password credential
          Username: test-user

        ID: test-credential-2
          Type: StringCredentialsImpl
          Scope: GLOBAL
          Description: Test secret text credential

    Returns:
        List[CredentialsDomain]: domains 列表，每個包含其 credentials
    """
    domains = []
    current_domain = None
    current_credential = None

    lines = stdout.split('\n')

    for line in lines:
        # Domain header
        domain_match = re.match(r'===\s*Domain:\s*(.+?)\s*===', line)
        if domain_match:
            if current_credential and current_domain:
                current_domain.credentials.append(current_credential)
            if current_domain:
                domains.append(current_domain)
            domain_name = domain_match.group(1)
            current_domain = CredentialsDomain(name=domain_name, credentials=[])
            current_credential = None
            continue

        # Credential ID (開始新的 credential)
        if line.startswith('ID:'):
            if current_credential and current_domain:
                current_domain.credentials.append(current_credential)

            cred_id = line.split(':', 1)[1].strip()
            current_credential = Credential(id=cred_id, type='')
            continue

        # Credential 屬性（以 2+ 空白開頭）
        if current_credential and line.startswith('  '):
            line = line.strip()
            if ':' in line:
                key, value = line.split(':', 1)
                value = value.strip()

                if key == 'Type':
                    current_credential.type = value
                elif key == 'Scope':
                    current_credential.scope = value
                elif key == 'Description':
                    current_credential.description = value
                elif key == 'Username':
                    current_credential.username = value

    # 加入最後一個 credential 和 domain
    if current_credential and current_domain:
        current_domain.credentials.append(current_credential)
    if current_domain:
        domains.append(current_domain)

    return domains


def find_credential_by_id(domains: List[CredentialsDomain], cred_id: str) -> Optional[Credential]:
    """在 domains 中尋找指定 ID 的 credential"""
    for domain in domains:
        for cred in domain.credentials:
            if cred.id == cred_id:
                return cred
    return None


@dataclass
class CredentialDetail:
    """Credential 詳細資訊（來自 describe-credentials）"""
    id: str
    type: str
    scope: str
    description: str
    domain: str
    # 可選的額外欄位
    username: Optional[str] = None
    # 安全性檢查：不應該包含實際的 secrets
    has_password: bool = False  # 僅標記是否有 password 欄位，但不儲存值
    has_secret: bool = False    # 僅標記是否有 secret 欄位，但不儲存值


def parse_credential_detail(stdout: str) -> CredentialDetail:
    """
    解析 describe-credentials 命令的輸出

    預期格式：
        === Domain: (global) ===

        ID: test-credential-1
        Type: UsernamePasswordCredentialsImpl
        Scope: GLOBAL
        Description: Test username/password credential

        Details:
          Username: test-user
          Password: ******** (hidden)

    Returns:
        CredentialDetail: credential 詳細資訊
    """
    lines = stdout.split('\n')

    cred_id = None
    cred_type = None
    scope = None
    description = None
    domain = None
    username = None
    has_password = False
    has_secret = False

    for line in lines:
        line = line.strip()

        # Domain header
        domain_match = re.match(r'===\s*Domain:\s*(.+?)\s*===', line)
        if domain_match:
            domain = domain_match.group(1)
            continue

        # 主要欄位
        if line.startswith('ID:'):
            cred_id = line.split(':', 1)[1].strip()
        elif line.startswith('Type:'):
            cred_type = line.split(':', 1)[1].strip()
        elif line.startswith('Scope:'):
            scope = line.split(':', 1)[1].strip()
        elif line.startswith('Description:'):
            description = line.split(':', 1)[1].strip()
        elif line.startswith('Username:'):
            username = line.split(':', 1)[1].strip()
        elif 'Password:' in line or 'password' in line.lower():
            # 檢查是否提到 password（但不擷取值）
            has_password = True
        elif 'Secret:' in line or 'secret' in line.lower():
            # 檢查是否提到 secret（但不擷取值）
            has_secret = True

    if not cred_id or not cred_type:
        raise ValueError("Failed to parse credential detail: missing required fields")

    return CredentialDetail(
        id=cred_id,
        type=cred_type,
        scope=scope or "UNKNOWN",
        description=description or "(no description)",
        domain=domain or "(unknown)",
        username=username,
        has_password=has_password,
        has_secret=has_secret
    )


# ============================================================================
# 測試函數 - List-Credentials 指令
# ============================================================================


def test_list_credentials_basic(run_jenkee_authed):
    """
    測試列出所有 Credentials

    對應 test plan 步驟 1
    """
    # Act: 執行 list-credentials 指令
    result = run_jenkee_authed.run("list-credentials")

    # Parse: 解析 credentials 列表
    domains = parse_credentials_list(result.stdout)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"list-credentials should succeed, got: {result.stderr}"
    assert len(domains) > 0, "Should have at least one domain"

    # 驗證包含測試 credentials
    cred1 = find_credential_by_id(domains, "test-credential-1")
    assert cred1 is not None, "Should contain test-credential-1"
    assert cred1.type == "UsernamePasswordCredentialsImpl", \
        f"test-credential-1 should be UsernamePasswordCredentialsImpl, got {cred1.type}"

    cred2 = find_credential_by_id(domains, "test-credential-2")
    assert cred2 is not None, "Should contain test-credential-2"
    assert cred2.type == "StringCredentialsImpl", \
        f"test-credential-2 should be StringCredentialsImpl, got {cred2.type}"

    cred3 = find_credential_by_id(domains, "test-credential-3")
    assert cred3 is not None, "Should contain test-credential-3"
    assert cred3.type == "UsernamePasswordCredentialsImpl", \
        f"test-credential-3 should be UsernamePasswordCredentialsImpl, got {cred3.type}"


def test_list_credentials_no_secrets_leaked(run_jenkee_authed):
    """
    測試 list-credentials 不洩漏 Secrets

    對應 test plan 步驟 2
    驗證輸出中不包含實際的 passwords, tokens, private keys
    """
    # Act: 執行 list-credentials 指令
    result = run_jenkee_authed.run("list-credentials")

    # Assert: 驗證不洩漏 secrets
    output = result.stdout.lower()

    # 不應該包含測試用的實際 secrets
    assert "test-password" not in output, \
        "Should not leak test-password"
    assert "admin-password" not in output, \
        "Should not leak admin-password"
    assert "test-secret-value" not in output, \
        "Should not leak test-secret-value"

    # 不應該有 "password=" 或 "secret=" 格式的洩漏
    assert re.search(r'password\s*=\s*\S+', output) is None, \
        "Should not have 'password=<value>' format"
    assert re.search(r'secret\s*=\s*\S+', output) is None, \
        "Should not have 'secret=<value>' format"
    assert re.search(r'token\s*=\s*\S+', output) is None, \
        "Should not have 'token=<value>' format"

    # 可以顯示 username（這不是 secret）
    # 但不應該同時顯示 password


def test_list_credentials_shows_metadata(run_jenkee_authed):
    """
    測試 list-credentials 顯示完整 Metadata

    驗證輸出包含必要的 metadata 欄位
    """
    # Act: 執行 list-credentials 指令
    result = run_jenkee_authed.run("list-credentials")

    # Parse: 解析 credentials 列表
    domains = parse_credentials_list(result.stdout)

    # Assert: 驗證每個 credential 都有必要的 metadata
    for domain in domains:
        for cred in domain.credentials:
            # 必要欄位
            assert cred.id, f"Credential should have ID"
            assert cred.type, f"Credential {cred.id} should have Type"
            assert cred.scope, f"Credential {cred.id} should have Scope"

            # Username 應該顯示（如果有的話）
            if cred.type == "UsernamePasswordCredentialsImpl":
                # UsernamePassword 類型應該有 username
                assert cred.username is not None, \
                    f"UsernamePassword credential {cred.id} should have username"


def test_list_credentials_multiple_types(run_jenkee_authed):
    """
    測試 list-credentials 正確處理不同類型的 Credentials

    對應 test plan 步驟 4
    """
    # Act: 執行 list-credentials 指令
    result = run_jenkee_authed.run("list-credentials")

    # Parse: 解析 credentials 列表
    domains = parse_credentials_list(result.stdout)

    # Assert: 驗證包含不同類型的 credentials
    all_creds = []
    for domain in domains:
        all_creds.extend(domain.credentials)

    # 取得所有類型
    types = {cred.type for cred in all_creds}

    # 應該至少有兩種類型
    assert len(types) >= 2, \
        f"Should have multiple credential types, got: {types}"

    # 應該包含測試用的兩種類型
    assert "UsernamePasswordCredentialsImpl" in types, \
        "Should have UsernamePasswordCredentialsImpl"
    assert "StringCredentialsImpl" in types, \
        "Should have StringCredentialsImpl"


# ============================================================================
# 測試函數 - Describe-Credentials 指令
# ============================================================================


def test_describe_credentials_basic(run_jenkee_authed):
    """
    測試查看特定 Credential 的詳細資訊

    對應 test plan 步驟 3
    """
    # Act: 執行 describe-credentials 指令
    result = run_jenkee_authed.run("describe-credentials", "test-credential-1")

    # Parse: 解析 credential 詳細資訊
    detail = parse_credential_detail(result.stdout)

    # Assert: 驗證執行成功
    assert result.returncode == 0, f"describe-credentials should succeed"
    assert detail.id == "test-credential-1", \
        f"Should describe test-credential-1, got {detail.id}"
    assert detail.type == "UsernamePasswordCredentialsImpl", \
        f"Should have correct type, got {detail.type}"


def test_describe_credentials_no_secrets_without_flag(run_jenkee_authed):
    """
    測試 describe-credentials 在沒有 --show-secret 時不洩漏 Secrets

    這是安全性的關鍵測試
    """
    # Act: 執行 describe-credentials 指令（不帶 --show-secret）
    result = run_jenkee_authed.run("describe-credentials", "test-credential-1")

    # Assert: 驗證不洩漏 secrets
    output = result.stdout.lower()

    # 不應該包含實際的 password
    assert "test-password" not in output, \
        "Should not leak test-password without --show-secret flag"

    # 測試 secret text credential
    result2 = run_jenkee_authed.run("describe-credentials", "test-credential-2")
    output2 = result2.stdout.lower()

    assert "test-secret-value" not in output2, \
        "Should not leak test-secret-value without --show-secret flag"


def test_describe_credentials_shows_username(run_jenkee_authed):
    """
    測試 describe-credentials 顯示 Username（非敏感資訊）

    Username 不是 secret，應該正常顯示
    """
    # Act: 執行 describe-credentials 指令
    result = run_jenkee_authed.run("describe-credentials", "test-credential-1")

    # Parse: 解析 credential 詳細資訊
    detail = parse_credential_detail(result.stdout)

    # Assert: 驗證顯示 username
    assert detail.username is not None, "Should show username"
    assert detail.username == "test-user", \
        f"Should show correct username, got {detail.username}"


def test_describe_credentials_different_types(run_jenkee_authed):
    """
    測試 describe-credentials 處理不同類型的 Credentials

    對應 test plan 步驟 4
    """
    # Test 1: UsernamePassword credential
    result1 = run_jenkee_authed.run("describe-credentials", "test-credential-1")
    detail1 = parse_credential_detail(result1.stdout)

    assert result1.returncode == 0, "Should succeed for UsernamePassword credential"
    assert detail1.type == "UsernamePasswordCredentialsImpl"
    assert detail1.username == "test-user"

    # Test 2: Secret text credential
    result2 = run_jenkee_authed.run("describe-credentials", "test-credential-2")
    detail2 = parse_credential_detail(result2.stdout)

    assert result2.returncode == 0, "Should succeed for StringCredentials"
    assert detail2.type == "StringCredentialsImpl"

    # Test 3: Another UsernamePassword credential
    result3 = run_jenkee_authed.run("describe-credentials", "test-credential-3")
    detail3 = parse_credential_detail(result3.stdout)

    assert result3.returncode == 0, "Should succeed for another UsernamePassword credential"
    assert detail3.type == "UsernamePasswordCredentialsImpl"
    assert detail3.username == "admin-user"


def test_describe_credentials_nonexistent(run_jenkee_authed):
    """
    測試查詢不存在的 Credential

    對應 test plan 錯誤情境測試
    """
    # Act: 執行 describe-credentials 指令（使用不存在的 ID）
    result = run_jenkee_authed.build_command(
        "describe-credentials", "non-existent-credential-id"
    ).allow_failure().run()

    # Assert: 驗證失敗
    assert result.returncode != 0, "Should fail for non-existent credential"
    error_output = (result.stdout + result.stderr).lower()
    assert 'error' in error_output or 'not found' in error_output, \
        f"Should have error message, got: {result.stdout + result.stderr}"


# ============================================================================
# 整合測試 - 完整工作流程
# ============================================================================


def test_complete_credentials_workflow(run_jenkee_authed):
    """
    測試完整的 credentials 查詢工作流程

    對應 test plan 場景 A：驗證環境 Credentials 設定
    """
    # 1. 列出所有 credentials
    list_result = run_jenkee_authed.run("list-credentials")
    domains = parse_credentials_list(list_result.stdout)

    assert list_result.returncode == 0, "Step 1: list-credentials should succeed"
    assert len(domains) > 0, "Step 1: Should have at least one domain"

    # 2. 檢查必要的 credentials 是否存在
    required_creds = ["test-credential-1", "test-credential-2", "test-credential-3"]

    for cred_id in required_creds:
        cred = find_credential_by_id(domains, cred_id)
        assert cred is not None, f"Step 2: {cred_id} should exist"

    # 3. 查看詳細資訊確認設定
    for cred_id in required_creds:
        desc_result = run_jenkee_authed.run("describe-credentials", cred_id)
        assert desc_result.returncode == 0, \
            f"Step 3: describe-credentials should succeed for {cred_id}"

        detail = parse_credential_detail(desc_result.stdout)
        assert detail.id == cred_id, \
            f"Step 3: Should describe {cred_id}, got {detail.id}"

    # 4. 驗證整個流程沒有洩漏 secrets
    all_output = list_result.stdout
    for cred_id in required_creds:
        desc_result = run_jenkee_authed.run("describe-credentials", cred_id)
        all_output += desc_result.stdout

    output_lower = all_output.lower()
    assert "test-password" not in output_lower, \
        "Step 4: Should not leak test-password in any output"
    assert "admin-password" not in output_lower, \
        "Step 4: Should not leak admin-password in any output"
    assert "test-secret-value" not in output_lower, \
        "Step 4: Should not leak test-secret-value in any output"


def test_credentials_output_is_parseable(run_jenkee_authed):
    """
    測試 credentials 指令的輸出可以被腳本解析

    確保輸出格式穩定，適合自動化使用
    """
    # Test list-credentials
    list_result = run_jenkee_authed.run("list-credentials")
    domains = parse_credentials_list(list_result.stdout)

    # 應該能夠成功解析
    assert len(domains) > 0, "Should be able to parse list-credentials output"

    # 每個 credential 應該有必要的欄位
    for domain in domains:
        for cred in domain.credentials:
            assert cred.id, "Parsed credential should have ID"
            assert cred.type, "Parsed credential should have Type"

    # Test describe-credentials
    desc_result = run_jenkee_authed.run("describe-credentials", "test-credential-1")
    detail = parse_credential_detail(desc_result.stdout)

    # 應該能夠成功解析
    assert detail.id == "test-credential-1", "Should be able to parse describe-credentials output"
    assert detail.type, "Parsed detail should have Type"



"""List builds command"""

import sys

from jenkins_tools.core import Command, JenkinsConfig, JenkinsCLI


class ListBuildsCommand(Command):
    """List build history for a Jenkins job"""

    def __init__(self, args):
        """
        Initialize with command line arguments

        Args:
            args: List of command arguments (sys.argv[2:])
        """
        self.args = args

    def execute(self) -> int:
        """Execute list-builds command"""
        config = JenkinsConfig()

        # Check if credentials are configured
        if not config.is_configured():
            print("Error: Jenkins credentials not configured.", file=sys.stderr)
            print(f"Run 'jks auth' to configure credentials.", file=sys.stderr)
            return 1

        # Parse arguments
        if not self.args:
            print("Error: Missing job name", file=sys.stderr)
            print("Usage: jks list-builds <job-name>", file=sys.stderr)
            return 1

        job_name = self.args[0]

        # Use groovy script to list builds
        cli = JenkinsCLI(config)
        groovy_script = f"""
def job = hudson.model.Hudson.instance.getItemByFullName('{job_name}')
if (job) {{
  job.builds.each {{ build ->
    def result = build.result ? build.result.toString() : 'BUILDING'
    def timestamp = build.time.format('yyyy-MM-dd HH:mm:ss')
    def duration = build.duration / 1000  // 轉換為秒
    println "#${{build.number}} ${{result}} ${{timestamp}} ${{duration}}s"
  }}
}} else {{
  println "ERROR: Job not found"
}}
"""

        result = cli.run("groovy", "=", stdin_input=groovy_script)

        if result.returncode == 0:
            output = result.stdout.strip()
            if output and not output.startswith("ERROR:"):
                print(output)
                return 0
            elif output.startswith("ERROR:"):
                print(f"Error: Job '{job_name}' not found", file=sys.stderr)
                return 1
            else:
                print(f"No builds found for job '{job_name}'")
                return 0
        else:
            print(f"Error: Failed to list builds for job '{job_name}'", file=sys.stderr)
            if result.stderr:
                print(result.stderr, file=sys.stderr)
            return 1
